import json
import logging
from json import JSONDecodeError
from pathlib import Path

import click
import pyperclip
import rich
from InquirerPy import inquirer
from rich.tree import Tree

from abst.bastion_support.oci_bastion import Bastion
from abst.cfg_func import __upgrade
from abst.config import default_contexts_location, share_excluded_keys
from abst.tools import get_context_path
from abst.utils.misc_funcs import get_context_data, setup_calls, get_context_set_data


@click.group(help="Contexts commands")
def context():
    pass


@click.group(help="Context commands alias")
def ctx():
    pass


@context.command("list", help="Will list all contexts in ~/.abst/context/ folder")
@click.option("--debug", is_flag=True, default=False)
def _list(debug=False):
    setup_calls(debug)

    tree = Tree("Contexts")

    for file in Path(default_contexts_location).iterdir():
        if file.name.startswith(".") or not file.name.endswith(".json"):
            continue
        cfg = Bastion.load_json(file)
        used_time = "" if "last-time-used" not in cfg.keys() else f"| Used: {cfg['last-time-used']}"
        tree.add(f"{file.name.replace('.json', '')} {used_time}")

    if not len(tree.children):
        rich.print("No contexts")

    rich.print(tree)


@context.command(help="Will display JSON format of context")
@click.option("--debug", is_flag=True, default=False)
@click.argument("name")
def display(name, debug=False):
    setup_calls(debug)
    data = get_context_data(name)
    if data is None:
        return
    rich.print_json(data=data)


@context.command(
    help="Will share context without local paths and put it in clipboard for sharing")
@click.option("--debug", is_flag=True, default=False)
@click.option("--raw", is_flag=True, default=False)
@click.argument("name")
def share(name: str, debug=False, raw=False):
    setup_calls(debug)

    if "/" in name:
        data = get_context_set_data(name)
    else:
        data = get_context_data(name)
    if data is None:
        return

    for key in share_excluded_keys:
        data.pop(key, None)
    data["default-name"] = "!YOUR NAME!"

    if not raw:
        rich.print(f"[bold]Context '{name}' config contents:[/bold]\n")
        logging.debug("Data transmitted into clipboard")
        pyperclip.copy(str(data))
        rich.print("Copied context into clipboard")
    rich.print_json(data=data)


@context.command(help="Will paste context from clipboard into provided context name")
@click.option("--debug", is_flag=True, default=False)
@click.argument("name")
def paste(name, debug=False):
    setup_calls(debug)

    data = pyperclip.paste()
    path = get_context_path(name)
    if data is None:
        return

    try:
        data_loaded = json.loads(data.replace("'", "\""))
    except JSONDecodeError:
        rich.print("[red]Invalid data, please try copying context again[/red]")
        exit(1)

    if data_loaded["default-name"] == "!YOUR NAME!":
        data_loaded["default-name"] = inquirer.text(
            "Please enter your name for bastion session that will be visible in OCI:").execute()

    if inquirer.confirm("Would you like to change IP?").execute():
        data_loaded["target-ip"] = inquirer.text(
            "Please enter your target IP address:").execute()

    if path.exists():
        if not inquirer.confirm("File exists. Overwrite?").execute():
            rich.print("[red]Canceled.[/red]")
            return

    Bastion.write_creds_json(data_loaded, path)
    rich.print(f"Wrote config into '{path}'")


@context.command("generate", help="Will generate sample json and overwrite changes")
@click.option("--debug", is_flag=True, default=False)
@click.argument("context-name", default=None, required=False)
def generate(debug, context_name):
    setup_calls(debug)

    path = get_context_path(context_name)

    Bastion.create_default_locations()
    td = Bastion.generate_sample_dict()
    creds_path = Bastion.write_creds_json(td, path)
    print(
        f"Sample credentials generated, please fill 'creds.json' in {creds_path} with "
        f"your credentials for this to work, you can use 'abst json fill "
        f"{context_name if context_name else ''}'"
    )


@context.command(
    "fill", help="Fills Json config with credentials you enter interactively"
)
@click.option("--debug", is_flag=True, default=False)
@click.argument("context-name", default=None, required=False)
def fill(debug, context_name):
    setup_calls(debug)

    path = get_context_path(context_name)

    if not path.exists():
        rich.print("Generating sample Creds file")
        Bastion.create_default_locations()
        td = Bastion.generate_sample_dict()
        Bastion.write_creds_json(td, path)

    if not default_contexts_location.exists():
        rich.print("Generating contexts location")
        Bastion.create_default_locations()

    rich.print(f"[green]Filling {str(path)}")
    rich.print("Please fill field one by one as displayed")
    n_dict = dict()

    creds_json_ro = Bastion.load_json(path)

    for key, value in creds_json_ro.items():
        n_dict[key] = inquirer.text(
            message=f"{key.capitalize()}:", default=value
        ).execute()
    rich.print("\n[red]New json looks like this:[/red]")
    rich.print_json(data=n_dict)
    if inquirer.confirm(message="Write New Json ?", default=False).execute():
        Bastion.write_creds_json(n_dict, path)
        rich.print("[green]Wrote changes[/green]")
    else:
        rich.print("[red]Fill interrupted, nothing changed[/red]")


@context.command(
    "locate", help="Locates Json config with credentials you enter interactively"
)
@click.option("--debug", is_flag=True, default=False)
@click.argument("context-name", default=None, required=False)
def locate(debug, context_name):
    setup_calls(debug)

    path = get_context_path(context_name)

    if path.exists():
        rich.print(f"[green]Config file location: {path.absolute()}[/green]")
    else:
        rich.print(
            f"[red]Config does not exist yet, future location"
            f" {path.absolute()}[/red]"
        )


@context.command(
    "upgrade", help="Upgrades Json config with credentials you enter interactively"
)
@click.option("--debug", is_flag=True, default=False)
@click.option("--all", is_flag=True, default=False)
@click.argument("context-name", default=None, required=False)
def upgrade(debug, context_name, all):
    setup_calls(debug)

    path = get_context_path(context_name)

    if not path.exists():
        rich.print("[green]No config to upgrade[/green]")
        return

    __upgrade(context_name, path, all)


ctx.add_command(_list, "list")
ctx.add_command(display, "display")
ctx.add_command(share, "share")
ctx.add_command(paste, "paste")
ctx.add_command(generate, "generate")
ctx.add_command(fill, "fill")
ctx.add_command(locate, "locate")
ctx.add_command(upgrade, "upgrade")



import signal
from pathlib import Path
from threading import Thread
from time import sleep
from typing import Optional

import rich
from click import clear
from rich import print
from rich.align import Align

from abst.bastion_support.oci_bastion import Bastion
from abst.config import default_stack_location, default_stack_contents, \
    default_contexts_location
from abst.wrappers import load_stack_decorator


class BastionScheduler:
    __dry_stack = set()
    __live_stack = set()
    stopped = False
    session_list = []

    @classmethod
    def load_stack(cls):
        if not default_stack_location.exists():
            Bastion.write_creds_json(default_stack_contents, default_stack_location)
        if default_stack_location.exists():
            cls.__dry_stack = set(Bastion.load_json(default_stack_location)["stack"])

    @classmethod
    def update_stack(cls):
        if default_stack_location.exists():
            Bastion.write_creds_json({"stack": list(cls.__dry_stack)},
                                     default_stack_location)

    @classmethod
    def get_contexts(cls):
        return [name.name.replace(".json", "") for name in
                default_contexts_location.iterdir()]

    @classmethod
    @load_stack_decorator
    def add_bastion(cls, context_name: str):
        if context_name in cls.__dry_stack:
            print("Already in stack")
            return

        if context_name != "default" and context_name not in cls.get_contexts():
            print(f"No context with name {context_name}")
            return

        if (res := cls.local_port_in_stack(context_name))[0]:
            rich.print(
                f"[red]Local port is already taken by {res[1].context_name}[/red]")
            return

        cls.__dry_stack.add(context_name)
        rich.print(f"[green]Added Bastion {context_name}[/green]")
        cls.update_stack()

    @classmethod
    @load_stack_decorator
    def remove_bastion(cls, name: str):
        try:
            cls.__dry_stack.remove(name)
            rich.print(f"[green]Removed Bastion {name}[/green]")
            cls.update_stack()
        except KeyError:
            rich.print(f"[red]No Bastion by name {name} in stack[/red]")

    @classmethod
    def __display(cls, console):
        console.clear()
        from rich.table import Table
        table = Table(title="Bastion Sessions", highlight=True)
        table.add_column("Name", justify="left", style="cyan", no_wrap=True)
        table.add_column("Local Port", style="magenta", no_wrap=True)
        table.add_column("Active", justify="right", style="green", no_wrap=True)
        table.add_column("Status", justify="right", style="green", no_wrap=True)
        for bastion in cls.__live_stack:
            conf = bastion.load_self_creds()
            active = bastion.connected and bastion.active_tunnel.poll() is None
            name = "default" if bastion.context_name is None else bastion.context_name

            try:
                table.add_row(name, conf.get('local-port', 'Not Specified'), str(active),
                              bastion.current_status)
            except:
                pass
        centered_table = Align.center(table)
        console.print(centered_table)

    @classmethod
    def __display_loop(cls):
        from rich.console import Console
        console = Console()
        signal.signal(signal.SIGINT, BastionScheduler.kill_all)
        signal.signal(signal.SIGTERM, BastionScheduler.kill_all)
        while True:
            clear()
            cls.__display(console)
            sleep(1.5)

    @classmethod
    def _run_indefinitely(cls, func, force: bool = False):
        while True:
            if cls.stopped:
                return
            try:
                func(force=force)
            finally:
                sleep(0.3)

    @classmethod
    @load_stack_decorator
    def run(cls, force=False, set_dir: Optional[Path] = None):
        signal.signal(signal.SIGINT, BastionScheduler.kill_all)
        signal.signal(signal.SIGTERM, BastionScheduler.kill_all)
        rich.print("Will run all Bastions in parallel")
        thread_list = []

        if not set_dir:
            for context_name in cls.__dry_stack:
                if cls.stopped:
                    return
                region = Bastion.load_json(Bastion.get_creds_path_resolve(context_name)).get("region", None)
                bastion = Bastion(None if context_name == "default" else context_name, region=region)
                Bastion.session_list.append(bastion)
                cls.session_list.append(bastion)
                cls.__live_stack.add(bastion)
                t = Thread(name=context_name, target=cls._run_indefinitely,
                           args=[bastion.create_forward_loop, force], daemon=True)
                thread_list.append(t)
                t.start()
                rich.print(f"Started {context_name}")
        else:
            for context_path in filter(lambda p: not str(p.name).startswith(".") and str(p.name).endswith(".json"),
                                       set_dir.iterdir()):
                if cls.stopped:
                    return
                context_name = context_path.name[:-5]
                region = Bastion.load_json(context_path).get("region", None)
                bastion = Bastion(None if context_name == "default" else context_name, region=region,
                                  direct_json_path=context_path)

                Bastion.session_list.append(bastion)
                cls.session_list.append(bastion)
                cls.__live_stack.add(bastion)
                t = Thread(name=context_name, target=cls._run_indefinitely,
                           args=[bastion.create_forward_loop, force], daemon=True)
                thread_list.append(t)
                t.start()
                rich.print(f"Started {context_name}")

        cls.__display_loop()

    @classmethod
    @load_stack_decorator
    def local_port_in_stack(cls, context_name: str):
        path = Bastion.get_creds_path_resolve(context_name)
        instance_creds = Bastion.load_json(path)
        for context_name in cls.__dry_stack:
            b_creds = Bastion.load_json(Bastion.get_creds_path_resolve(context_name))
            if b_creds["local-port"] == instance_creds["local-port"]:
                return True, context_name
        return False, None

    @classmethod
    def kill_all(cls, a=None, b=None, c=None):
        # This should be only executed in running state
        cls.stopped = True
        blist_copy = list(cls.session_list)
        for i, sess in enumerate(blist_copy):
            try:
                rich.print(f"[red]Killing[/red] {sess.bid}")
                Bastion.current_status = "deleting"
                sess.kill()
                Bastion.delete_bastion_session(sess.bid, sess.region)
            except Exception:
                pass
            finally:
                print(f"Deleting {i + 1}/{len(blist_copy)}")
        exit(0)

    @classmethod
    @load_stack_decorator
    def get_bastions(cls) -> tuple:
        return tuple(cls.__dry_stack)



import json
import logging
import os
import subprocess
from pathlib import Path
from threading import Thread
from time import sleep
from typing import Optional

import rich
from rich.logging import RichHandler

from abst.bastion_support.oci_bastion import Bastion
from abst.config import default_contexts_location, default_parallel_sets_location


def setup_calls(debug):
    setup_debug(debug)


def setup_debug(debug):
    if not debug:
        logging.disable(logging.DEBUG)
        logging.disable(logging.INFO)
    from imp import reload  # python 2.x don't need to import reload, use it directly
    reload(logging)
    logging.basicConfig(
        level=logging.DEBUG if debug else logging.CRITICAL,
        format="%(message)s",
        datefmt="[%X]",
        handlers=[RichHandler()],
    )


def print_eligible(searched: str):
    contexts = Bastion.get_contexts()
    rich.print("[bold]Configured contexts:[/bold]")
    for key, context in contexts.items():
        do_skip = False
        for value in context.values():
            if searched in value:
                do_skip = True
                break
        if do_skip:
            continue
        rich.print(key)


def get_context_data(name) -> Optional[dict]:
    normalized_name = name if ".json" in name else name + ".json"
    if normalized_name in [file.name for file in
                           Path(default_contexts_location).iterdir()]:
        with open(Path(default_contexts_location) / normalized_name, "r") as f:
            data = json.load(f)
            return data
    else:
        rich.print("[red]Context does not exists[/red]")
        return None


def get_context_set_data(name) -> Optional[dict]:
    path = name.split("/")
    if len(path) != 2:
        rich.print(f"[red]Invalid path '{name}'[/red]")

    set_dir_name = path[0]
    if set_dir_name in [file.name for file in
                        Path(default_parallel_sets_location).iterdir()]:
        normalized_name = path[1] if ".json" in path[1] else path[1] + ".json"
        if normalized_name in [file.name for file in
                               Path(default_parallel_sets_location / set_dir_name).iterdir()]:
            with open(
                    Path(default_parallel_sets_location) / set_dir_name / normalized_name,
                    "r") as f:
                data = json.load(f)
                return data
        else:
            rich.print("[red]Context does not exists[/red]")
            return None
    else:
        rich.print("[red]Set does not exists[/red]")
        return None


def fetch_pods():
    rich.print("Fetching pods")
    pod_lines = (
        subprocess.check_output(f"kubectl get pods -A".split(" "))
        .decode()
        .split("\n")
    )
    return pod_lines


def recursive_copy(dir_to_iter: Path, dest_path: str, exclude: str, data: list,
                   pod_name_precise: str,
                   thread_list: list):
    if dir_to_iter.is_dir():
        create_folder_kubectl(data, dest_path, pod_name_precise)

    for file in dir_to_iter.iterdir():
        final_dest_path = make_final_dest_path(dest_path, file)
        if file.name == exclude:
            rich.print(f"   [yellow]{file} excluded.[/yellow]")
            continue
        elif file.is_dir():
            rich.print(f"[yellow]Recursion to folder {file}[/yellow]")

            create_folder_kubectl(data, final_dest_path, pod_name_precise)
            t = Thread(name=f"recursive_copy_{file}", target=recursive_copy,
                       args=[file, dest_path, exclude, data, pod_name_precise,
                             thread_list])
            t.start()
            thread_list.append(t)
        else:
            copy_file_alt_kubectl(data, final_dest_path, file, pod_name_precise)


def make_final_dest_path(dest_path, file):
    sub_path = str(file).split("/")

    fsp = sub_path[sub_path.index(Path(dest_path).name) + 1:]

    final_dest_path = f"{dest_path}{'/' if len(fsp) > 0 else ''}{'/'.join(fsp)}"
    return final_dest_path


def create_folder_kubectl(data: list, dest_path: str, pod_name_precise: str):
    kubectl_create_dir_cmd = (f"kubectl"
                              f" exec -i {pod_name_precise} -n {data[0]} -- mkdir -p \"{dest_path}\"")
    logging.info(f"Executing {kubectl_create_dir_cmd}")
    os.system(kubectl_create_dir_cmd)


def copy_file_alt_kubectl(data: list, dest_path: str, local_path: Path,
                          pod_name_precise: str, tries=4):
    kubectl_alt_copy_cmd = (f"cat \"{local_path}\" |"
                            f" kubectl exec -i {pod_name_precise} -n {data[0]} -- tee \"{dest_path}\" > /dev/null")
    logging.info(f"Executing {kubectl_alt_copy_cmd}")
    exit_code = subprocess.call(kubectl_alt_copy_cmd, shell=True)
    rich.print(
        f"   [green]{local_path if type(local_path) == str else local_path.name}"
        f" ({'?' if type(local_path) == str else local_path.stat().st_size / 1000} kB)"
        f" successfully copied![/green]")
    if exit_code != 0 and tries < 0:
        rich.print("[red]Failed to copy.[/red]")
        exit(1)
    elif exit_code != 0 and tries > 0:
        rich.print(f"Failed to copy will try again, try {tries}/4")
        sleep(5)
        copy_file_alt_kubectl(data, dest_path, local_path, pod_name_precise, tries - 1)



import os

import rich


def filter_keys_by_substring(dictionary: dict, substring: str) -> list:
    # Convert substring to lower a case for case-insensitive comparison
    substring_lower = substring.lower()
    # Use a comprehension to filter keys
    filtered_keys = [key for key in dictionary.keys() if substring_lower in key.lower()]
    return filtered_keys


def filter_keys_by_port(data: dict, port: int):
    """Filter and return keys from the servers dictionary where the port matches the given input."""
    return [key for key, data in data.items() if port in str(data.get('port', None))]


def do_ssh(context_name: str, username, port):
    rich.print(f"[green]Running SSH to {context_name}[/green] "
               f"[yellow]{username}@localhost:{port}[/yellow]")
    os.system(f'ssh {username}@127.0.0.1 -p {port}')



from pathlib import Path

# Default config
default_stack_location: Path = (Path().home().resolve() / ".abst" / "stack.json")
default_conf_path: Path = (Path().home().resolve() / ".abst" / "config.json")
default_creds_path: Path = (Path().home().resolve() / ".abst" / "creds.json")
default_contexts_location: Path = (Path().home().resolve() / ".abst" / "contexts")
default_parallel_sets_location: Path = (Path().home().resolve() / ".abst" / "sets")

default_context_keys: tuple = (
    "host", "bastion-id", "default-name", "ssh-pub-path", "private-key-path", "target-ip",
    "local-port", "target-port", "ttl", "resource-id", "resource-os-username", "region",
    "ssh-custom-arguments")
default_stack_contents: dict = {"stack": []}
default_conf_contents: dict = {"used_context": None, "private-key-path": "~/.ssh/id_rsa",
                               "ssh-pub-path": "~/.ssh/id_rsa"
                                               ".pub"}

# Share Config
share_excluded_keys: tuple = ("private-key-path", "ssh-pub-path", "last-time-used")

# Max Shared JSON size
max_json_shared = 1048576  # Bytes
broadcast_shm_name = "abst_shared_memory"


def get_public_key(ssh_path):
    with open(str(Path(ssh_path).expanduser().resolve()), "r") as f:
        return f.read()



"""
__version__.py
~~~~~~~~~~~~~~

Information about the current version of the py-package-template package.
"""

__title__ = "abst"
__description__ = (
    "CLI Command making OCI Bastion and kubernetes usage simple and fast"
)

__version__ = "2.3.48"
__author__ = "Jiri Otoupal"
__author_email__ = "jiri-otoupal@ips-database.eu"
__license__ = "MIT"
__url__ = "https://github.com/jiri-otoupal/abst"
__pypi_repo__ = "https://pypi.org/project/abst/"

__version_name__ = "SSHed Penguin"
__change_log__ = """
* 'abst ssh' will show screen with available instances for ssh from all abst processes\n
* moved old ssh to pod from 'abst ssh' to 'abst pod ssh'
"""



from time import sleep

import click
import rich

from abst.bastion_support.oci_bastion import Bastion
from abst.utils.misc_funcs import setup_calls, print_eligible


@click.group(help="Group of commands for creating Bastion sessions")
def create():
    pass


@click.group(name="do", help="Alias Group of commands for creating Bastion sessions")
def _do():
    pass


@create.command(
    "forward",
    help="Creates and connects to Bastion session indefinitely until terminated by user",
)
@click.option("--shell", is_flag=True, default=False)
@click.option("--debug", is_flag=True, default=False)
@click.argument("context-name", default=None, required=False)
def forward(shell, debug, context_name):
    """Creates and connects to bastion sessions
    automatically until terminated"""

    setup_calls(debug)

    if context_name == "?":
        print_eligible("required only for Port Forward session")
        return

    if context_name is None:
        conf = Bastion.load_config()
        used_name = conf["used_context"]
    else:
        used_name = context_name
    try:
        while True:
            Bastion(used_name, region=
            Bastion.load_json(Bastion.get_creds_path_resolve(context_name)).get("region",
                                                                                None)).create_forward_loop(
                shell=shell)

            sleep(1)
    except FileNotFoundError:
        rich.print("[red]No such context found[/red]")


@create.command(
    "managed",
    help="Creates and connects to Bastion session indefinitely until terminated by user",
)
@click.option("--shell", is_flag=True, default=False)
@click.option("--debug", is_flag=True, default=False)
@click.argument("context-name", default=None, required=False)
def managed(shell, debug, context_name):
    """Creates and connects to bastion sessions
    automatically until terminated"""
    setup_calls(debug)

    if context_name == "?":
        print_eligible("required only for Managed SSH session")
        return

    if context_name is None:
        conf = Bastion.load_config()
        used_name = conf["used_context"]
    else:
        used_name = context_name

    try:
        while True:
            Bastion(used_name, region=
            Bastion.load_json(Bastion.get_creds_path_resolve(context_name)).get("region",
                                                                                None)).create_managed_loop(
                shell=shell)

            sleep(1)
    except FileNotFoundError:
        rich.print("[red]No such context found[/red]")


_do.add_command(forward)
_do.add_command(managed)
